/**
 * <copyright>
 * 
 * Copyright (c) Continental AG and others.
 * All rights reserved. This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 
 * which accompanies this distribution, and is
 * available at http://www.eclipse.org/org/documents/epl-v10.php
 * 
 * Contributors: 
 *     Continental AG - Initial API and implementation
 * 
 * </copyright>
 */

package eaadapter.tests;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.util.Arrays;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.junit.BeforeClass;
import org.junit.Test;

import eaadapter.EARepository;
import eaadapter.EaadapterFactory;
import eaadapter.common.util.ModelConstants;
import eaadapter.datatypes.MetaModelType;
import eaadapter.ea2ecore.EA2Ecore;
import eaadapter.ea2ecore.postprocessings.AddOppositeReferenceForGEASTADL;
import eaadapter.ea2ecore.postprocessings.ConfigureInstanceRefAssociation;
import eaadapter.ea2ecore.postprocessings.ConfigureInstanceRefClass;
import eaadapter.ea2ecore.postprocessings.ConvertEClassPrimitivesToEDataTypes;
import eaadapter.ea2ecore.postprocessings.DeleteEmptyIdentifiablePackage;
import eaadapter.ea2ecore.postprocessings.DeletePackageGeneratedEcore;
import eaadapter.ea2ecore.postprocessings.MakeAllFeaturesOrdered;
import eaadapter.ea2ecore.postprocessings.MakeEEnumEDataTypeUnsettable;
import eaadapter.ea2ecore.postprocessings.NameChanger;
import eaadapter.ea2ecore.postprocessings.NormalizePackageNsURI;
import eaadapter.ea2ecore.postprocessings.PackageNamesToLowercase;
import eaadapter.ea2ecore.postprocessings.SetAttributesDefaultValue;
import eaadapter.ea2ecore.postprocessings.SetInstanceClassNameForDataTypes;
import eaadapter.util.Util;

public class TestEcoreTransformation {
	
	private static EARepository eaModel;
	private static File eaModelFile;
	private static final String filename = "resources/testmodel.eap";
	private static final String ecoreFilename = "gen/testmodel.ecore";
	private static EPackage generatedEcore;
	
	@BeforeClass
	public static void initializeEAModel() throws Exception {
		eaModelFile = new File(filename);
		assertTrue("testmodel.eap exists", eaModelFile.exists());
		eaModel = EaadapterFactory.eINSTANCE.createEARepository();
		assertNotNull("eaModel must not be null", eaModel);
		eaModel.setFile(eaModelFile);
		eaModel.setTypeOfMetaModel(MetaModelType.EASTADL);
		eaModel.load();
		
		assertTrue("Model contains more than 1 element", eaModel.getModels().size() > 0);	

        EARepository copy = EcoreUtil.copy(eaModel);
        EA2Ecore ecoreGenerator = new EA2Ecore(copy);
        ecoreGenerator.setMetaModelType(ModelConstants.METAMODEL_TYPE_EASTADL);
        ecoreGenerator.getPostProcessings().add(new DeletePackageGeneratedEcore());
    	ecoreGenerator.getPostProcessings().add(new AddOppositeReferenceForGEASTADL());        
        ecoreGenerator.getPostProcessings().add(new NameChanger());
        ecoreGenerator.getPostProcessings().add(new ConvertEClassPrimitivesToEDataTypes());
        ecoreGenerator.getPostProcessings().add(new SetInstanceClassNameForDataTypes());       
        ecoreGenerator.getPostProcessings().add(new NormalizePackageNsURI("2.1.10"));
        ecoreGenerator.getPostProcessings().add(new MakeAllFeaturesOrdered());
        ecoreGenerator.getPostProcessings().add(new MakeEEnumEDataTypeUnsettable());
        ecoreGenerator.getPostProcessings().add(new SetAttributesDefaultValue());
        ecoreGenerator.getPostProcessings().add(new ConfigureInstanceRefAssociation());
        ecoreGenerator.getPostProcessings().add(new ConfigureInstanceRefClass());
        ecoreGenerator.getPostProcessings().add(new DeleteEmptyIdentifiablePackage());
        ecoreGenerator.getPostProcessings().add(new PackageNamesToLowercase());		
        ecoreGenerator.generateEcore();
        generatedEcore = ecoreGenerator.getEcore();
        eaadapter.ea2ecore.Util.saveModel(ecoreGenerator.getEcore(), new File(ecoreFilename));		
	}


	@Test
	public void testRootPackage() {
		assertNotNull(generatedEcore);
		assertEquals("rootPackage is 'DomainModel'", generatedEcore.getName(), "DomainModel");
		assertEquals("'DomainModel has 2 subpackages'", generatedEcore.getESubpackages().size(), 2);
		assertEquals("Check nsURI of DomainModel", generatedEcore.getNsURI(), "http://east-adl.info/2.1.10/DomainModel");
		assertEquals("nsPrefix is empty", generatedEcore.getNsPrefix(), "");
		EList<EAnnotation> annotations = generatedEcore.getEAnnotations();
		assertEquals(annotations.size(), 1);
		EAnnotation metadata = annotations.get(0);
		assertEquals(metadata.getSource(), "MetaData");
	}
	
	@Test
	public void testSystemModeling() {
		EPackage systemModelPackage = Util.findEPackageByName("systemmodeling", generatedEcore);
		assertNotNull(systemModelPackage);
		assertEquals(systemModelPackage.getName(), "systemmodeling");
		EList<EAnnotation> annotations = systemModelPackage.getEAnnotations();
		assertTrue("We must have 2 Annotations", annotations.size() == 2);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		assertEquals(annotations.get(1).getSource(), "http://www.eclipse.org/emf/2002/GenModel");
		
		assertEquals("systemmodeling package must have 5 classes", systemModelPackage.getEClassifiers().size(), 5);
		EClass systemModel = (EClass) systemModelPackage.getEClassifier("SystemModel");
		assertNotNull(systemModel);
		assertTrue("SystemModel has 1 supertype", systemModel.getESuperTypes().size() == 1);
		EReference vehicleLevelReference = (EReference) systemModel.getEStructuralFeature("vehicleLevel");
		assertNotNull(vehicleLevelReference);
		EClass vehicleLevel = (EClass) systemModelPackage.getEClassifier("VehicleLevel");
		assertNotNull(vehicleLevel);
		assertTrue("reference vehicleLevel references the VehicleLevel class", vehicleLevelReference.getEType() == vehicleLevel);
		assertEquals(vehicleLevelReference.isContainment(), true);
		assertEquals(vehicleLevelReference.isDerived(), false);
		assertEquals(vehicleLevelReference.getLowerBound(), 0);
		assertEquals(vehicleLevelReference.getUpperBound(), 1);
	}
	
	@Test
	public void testFunctionModeling() {
		EPackage functionmodelPackage = Util.findEPackageByName("functionmodeling", generatedEcore);
		assertNotNull(functionmodelPackage);
		assertEquals(functionmodelPackage.getName(), "functionmodeling");
		EList<EAnnotation> annotations = functionmodelPackage.getEAnnotations();
		assertTrue("We must have 2 Annotations", annotations.size() == 2);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		assertEquals(annotations.get(1).getSource(), "http://www.eclipse.org/emf/2002/GenModel");
		
		assertEquals("functionmodeling package must have 23 classes", functionmodelPackage.getEClassifiers().size(), 23);
		
		EClass allocateableElement = (EClass) functionmodelPackage.getEClassifier("AllocateableElement");
		assertNotNull(allocateableElement);
		
		EClass Allocation = (EClass) functionmodelPackage.getEClassifier("Allocation");
		assertNotNull(Allocation);
		assertTrue("Allocation has 1 supertype", Allocation.getESuperTypes().size() == 1);

		EClass AnalysisFunctionPrototype = (EClass) functionmodelPackage.getEClassifier("AnalysisFunctionPrototype");
		assertNotNull(AnalysisFunctionPrototype);
		assertTrue("AnalysisFunctionPrototype has 1 supertype", AnalysisFunctionPrototype.getESuperTypes().size() == 1);

		EClass AnalysisFunctionType = (EClass) functionmodelPackage.getEClassifier("AnalysisFunctionType");
		assertNotNull(AnalysisFunctionType);
		assertTrue("AnalysisFunctionType has 2 supertype", AnalysisFunctionType.getESuperTypes().size() == 1);
		
		EClass BasicSoftwareFunctionType = (EClass) functionmodelPackage.getEClassifier("BasicSoftwareFunctionType");
		assertNotNull(BasicSoftwareFunctionType);
		assertTrue("BasicSoftwareFunctionType has 1 supertype", BasicSoftwareFunctionType.getESuperTypes().size() == 1);
		
		EClass DesignFunctionPrototype = (EClass) functionmodelPackage.getEClassifier("DesignFunctionPrototype");
		assertNotNull(DesignFunctionPrototype);
		assertTrue("DesignFunctionPrototype has 2 supertype", DesignFunctionPrototype.getESuperTypes().size() == 2);
		
		EClass DesignFunctionType = (EClass) functionmodelPackage.getEClassifier("DesignFunctionType");
		assertNotNull(DesignFunctionType);
		assertTrue("DesignFunctionType has 1 supertype", DesignFunctionType.getESuperTypes().size() == 1);
		
		EClass FunctionalDevice = (EClass) functionmodelPackage.getEClassifier("FunctionalDevice");
		assertNotNull(FunctionalDevice);
		assertTrue("FunctionalDevice has 1 supertype", FunctionalDevice.getESuperTypes().size() == 1);
		
		EClass FunctionAllocation = (EClass) functionmodelPackage.getEClassifier("FunctionAllocation");
		assertNotNull(FunctionAllocation);
		assertTrue("FunctionAllocation has 1 supertype", FunctionAllocation.getESuperTypes().size() == 1);
		
		EClass FunctionClientServerInterface = (EClass) functionmodelPackage.getEClassifier("FunctionClientServerInterface");
		assertNotNull(FunctionClientServerInterface);
		assertTrue("FunctionClientServerInterface has 1 supertype", FunctionClientServerInterface.getESuperTypes().size() == 1);
		

		EClass FunctionClientServerPort = (EClass) functionmodelPackage.getEClassifier("FunctionClientServerPort");
		assertNotNull(FunctionClientServerPort);
		assertTrue("FunctionClientServerPort has 1 supertype", FunctionClientServerPort.getESuperTypes().size() == 1);
		
		EClass FunctionConnector = (EClass) functionmodelPackage.getEClassifier("FunctionConnector");
		assertNotNull(FunctionConnector);
		assertTrue("FunctionConnector has 3 supertype", FunctionConnector.getESuperTypes().size() == 3);
		
		EClass FunctionFlowPort = (EClass) functionmodelPackage.getEClassifier("FunctionFlowPort");
		assertNotNull(FunctionFlowPort);
		assertTrue("FunctionFlowPort has 1 supertype", FunctionFlowPort.getESuperTypes().size() == 1);
		
		EClass FunctionPort = (EClass) functionmodelPackage.getEClassifier("FunctionPort");
		assertNotNull(FunctionPort);
		assertTrue("FunctionPort has 2 supertype", FunctionPort.getESuperTypes().size() == 2);
		
		EClass FunctionPowerPort = (EClass) functionmodelPackage.getEClassifier("FunctionPowerPort");
		assertNotNull(FunctionPowerPort);
		assertTrue("FunctionPowerPort has 1 supertype", FunctionPowerPort.getESuperTypes().size() == 1);
		
		EClass FunctionPrototype = (EClass) functionmodelPackage.getEClassifier("FunctionPrototype");
		assertNotNull(FunctionPrototype);
		assertTrue("FunctionPrototype has 2 supertype", FunctionPrototype.getESuperTypes().size() == 2);
		
		EClass FunctionType = (EClass) functionmodelPackage.getEClassifier("FunctionType");
		assertNotNull(FunctionType);
		assertTrue("FunctionType has 2 supertype", FunctionType.getESuperTypes().size() == 2);
		
		EClass HardwareFunctionType = (EClass) functionmodelPackage.getEClassifier("HardwareFunctionType");
		assertNotNull(HardwareFunctionType);
		assertTrue("HardwareFunctionType has 1 supertype", HardwareFunctionType.getESuperTypes().size() == 1);
		
		EClass LocalDeviceManager = (EClass) functionmodelPackage.getEClassifier("LocalDeviceManager");
		assertNotNull(LocalDeviceManager);
		assertTrue("LocalDeviceManager has 1 supertype", LocalDeviceManager.getESuperTypes().size() == 1);
		
		EClass Operation = (EClass) functionmodelPackage.getEClassifier("Operation");
		assertNotNull(Operation);
		assertTrue("Operation has 1 supertype", Operation.getESuperTypes().size() == 1);
		
		EClass PortGroup = (EClass) functionmodelPackage.getEClassifier("PortGroup");
		assertNotNull(PortGroup);
		assertTrue("PortGroup has 1 supertype", PortGroup.getESuperTypes().size() == 1);
		
		EReference portReference = (EReference) PortGroup.getEStructuralFeature("port");
		assertNotNull(portReference);
		assertEquals(portReference.isContainer(), false);
		assertEquals(portReference.isContainment(), false);
		assertEquals(portReference.isDerived(), false);
		assertEquals(portReference.getEOpposite(), null);
		assertEquals(portReference.getEType(), FunctionPort);
		assertEquals(portReference.getLowerBound(), 0);
		assertEquals(portReference.getUpperBound(), -1);
		assertEquals(portReference.isTransient(), false);
		assertEquals(portReference.isVolatile(), false);
	}
	
	
	@Test
	public void testElements() {
		EPackage elementsPackage = Util.findEPackageByName("elements", generatedEcore);
		assertNotNull(elementsPackage);
		assertEquals(elementsPackage.getName(), "elements");
		assertEquals("elementsPackage must have 14 classes", elementsPackage.getEClassifiers().size(), 14);
		
		EClass Comment = (EClass) elementsPackage.getEClassifier("Comment");
		assertNotNull(Comment);
		EList<EAttribute> coomentattrib = Comment.getEAttributes();
		EAttribute attribute = coomentattrib.get(0);
		assertEquals(attribute.getName(), "body");
		
		EClass TraceableSpecification = (EClass) elementsPackage.getEClassifier("TraceableSpecification");
		assertNotNull(TraceableSpecification);
		EList<EAttribute> TraceableSpecificationattrib = TraceableSpecification.getEAttributes();
		EAttribute b = TraceableSpecificationattrib.get(0);
		assertEquals(b.getName(),"text");
		
	}
	
	@Test
	public void testDatatypes() {
		EPackage datatypesPackage = Util.findEPackageByName("datatypes", generatedEcore);
		assertNotNull(datatypesPackage);
		assertEquals(datatypesPackage.getName(), "datatypes");

		EClass ArrayDatatype = (EClass) datatypesPackage.getEClassifier("ArrayDatatype");
		assertNotNull(ArrayDatatype);
		EList<EAttribute> ArrayDatatypeattrib = ArrayDatatype.getEAttributes();
		List<String> attributeNames = Arrays.asList(new String[]{"maxLength", "minLength"});
		EAttribute e = ArrayDatatypeattrib.get(0);
		assertTrue(attributeNames.contains(e.getName()));
		EAttribute t = ArrayDatatypeattrib.get(1);
		assertTrue(attributeNames.contains(t.getName()));		

		EClass Enumeration = (EClass) datatypesPackage.getEClassifier("Enumeration");
		assertNotNull(Enumeration);
		EList<EAttribute> Enumerationattrib = Enumeration.getEAttributes();
		EAttribute p = Enumerationattrib.get(0);
		assertEquals(p.getName(),"isMultiValued");
		
		EClass Quantity = (EClass) datatypesPackage.getEClassifier("Quantity");
		assertNotNull(Quantity);
		List<String> quantityAttributeNames = Arrays.asList(new String[]{"timeExp", "thermodynamicTemperatureExp", "massExp", "amountOfSubstanceExp", "electricCurrentExp", "luminousIntensityExp", "lengthExp"});
		for (EAttribute attribute : Quantity.getEAttributes()) {
			assertTrue(quantityAttributeNames.contains(attribute.getName()));
		}
	}
	
	@Test
	public void testValues() {
		EPackage valuesPackage = Util.findEPackageByName("values", generatedEcore);
		assertNotNull(valuesPackage);
		assertEquals(valuesPackage.getName(), "values");
		EList<EAnnotation> annotations = valuesPackage.getEAnnotations();
		assertTrue("We must have 1 Annotations", annotations.size() == 1);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		assertEquals("valuesPackage must have 8 classes", valuesPackage.getEClassifiers().size(), 8);
		
		EClass EABooleanValue = (EClass) valuesPackage.getEClassifier("EABooleanValue");
		assertNotNull(EABooleanValue);
		EABooleanValue.getEAllAttributes();
		EList<EAttribute> ArrayDatatypeattrib = EABooleanValue.getEAttributes();
		EAttribute e = ArrayDatatypeattrib.get(0);
		assertEquals(e.getName(), "value");
		
		EClass EANumericalValue = (EClass) valuesPackage.getEClassifier("EANumericalValue");
		assertNotNull(EANumericalValue);
		EANumericalValue.getEAllAttributes();
		EList<EAttribute> EANumericalValueattrib = EANumericalValue.getEAttributes();
		EAttribute ea1 = EANumericalValueattrib.get(0);
		assertEquals(ea1.getName(), "value");
		
		EClass EAStringValue = (EClass) valuesPackage.getEClassifier("EAStringValue");
		assertNotNull(EAStringValue);
		EAStringValue.getEAllAttributes();
		EList<EAttribute> EAStringValueattrib = EAStringValue.getEAttributes();
		EAttribute ea2 = EAStringValueattrib.get(0);
		assertEquals(ea2.getName(), "value");		
	}
	
	@Test
	public void testUserattributes() {
		EPackage userattributesPackage = Util.findEPackageByName("userattributes", generatedEcore);
		assertNotNull(userattributesPackage);
		assertEquals(userattributesPackage.getName(), "userattributes");
		EList<EAnnotation> annotations = userattributesPackage.getEAnnotations();
		assertTrue("We must have 2 Annotations", annotations.size() == 2);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		assertEquals(annotations.get(1).getSource(), "http://www.eclipse.org/emf/2002/GenModel");
		
		assertEquals("useratrributesPackage must have 3 classes", userattributesPackage.getEClassifiers().size(), 3);
	}
	
	@Test
	public void testTimeValue() {
		EPackage timevaluePackage = Util.findEPackageByName("timevalue", generatedEcore);
		assertNotNull(timevaluePackage);
		assertEquals(timevaluePackage.getName(), "timevalue");
		EList<EAnnotation> annotations = timevaluePackage.getEAnnotations();
		assertTrue("We must have 1 Annotations", annotations.size() == 1);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		
		assertEquals("timevalue package must have 2 class", timevaluePackage.getEClassifiers().size(), 2);
		
		EClass TimeValue = (EClass) timevaluePackage.getEClassifier("TimeValue");
		assertNotNull(TimeValue);
		TimeValue.getEAllAttributes();
		EList<EAttribute> TimeValueattrib = TimeValue.getEAttributes();
		assertEquals(TimeValueattrib.size(), 2);
	}
	
	@Test
	public void testIdentifiable() {
		EPackage identifiablePackage = Util.findEPackageByName("identifiable", generatedEcore);
		assertNotNull(identifiablePackage);
		assertEquals(identifiablePackage.getName(), "identifiable");
		EList<EAnnotation> annotations = identifiablePackage.getEAnnotations();
		assertTrue("We must have 1 Annotations", annotations.size() == 1);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		
		assertEquals("identifiablePackage must have 2 class", identifiablePackage.getEClassifiers().size(), 2);
		
		EClass Referrable = (EClass) identifiablePackage.getEClassifier("Referrable");
		assertNotNull(Referrable);
		Referrable.getEAllAttributes();
		EList<EAttribute> Referrableattrib = Referrable.getEAttributes();
		EAttribute ea2 = Referrableattrib.get(0);
		assertEquals(ea2.getName(), "shortName");
	}
	
	@Test
	public void testSystemTemplate() {
		EPackage systemtemplatePackage = Util.findEPackageByName("systemtemplate", generatedEcore);
		assertNotNull(systemtemplatePackage);
		assertEquals(systemtemplatePackage.getName(), "systemtemplate");
		EList<EAnnotation> annotations = systemtemplatePackage.getEAnnotations();
		assertTrue("We must have 1 Annotations", annotations.size() == 1);
		assertEquals(annotations.get(0).getSource(), "MetaData");
		
		assertEquals("systemtemplatePackage must have 1 class", systemtemplatePackage.getEClassifiers().size(), 1);
		
		EClass System = (EClass) systemtemplatePackage.getEClassifier("System");
		assertNotNull(System);
		System.getEAllAttributes();
		EList<EAttribute> Systemattrib = System.getEAttributes();
		assertEquals(Systemattrib.size(), 2);
	}
	

	@Test
	public void testEnumeration() {
		EPackage functionModelingPackage = Util.findEPackageByName("functionmodeling", generatedEcore);
		assertNotNull(functionModelingPackage);
		EClassifier EADirectionKind = functionModelingPackage.getEClassifier("EADirectionKind");
		assertNotNull(EADirectionKind);
		assertTrue(EADirectionKind instanceof EEnum);	
		EList<EEnumLiteral> literals = ((EEnum) EADirectionKind).getELiterals();
		assertTrue(literals.size() == 3);
		List<String> literalValues = Arrays.asList("IN", "OUT", "INOUT");
		for (EEnumLiteral literal : literals) {
			assertTrue(literalValues.contains(literal.getLiteral()));
		}
	}
	 
	@Test
	public void testPrimiviveTypes() {
		EPackage primtiveTypesPackage = Util.findEPackageByName("primitivetypes", generatedEcore);
		assertNotNull(primtiveTypesPackage);	
		EClassifier Address = primtiveTypesPackage.getEClassifier("Address");
		assertNotNull(Address);
		assertTrue(Address instanceof EDataType);
		assertEquals(((EDataType)Address).getInstanceClassName(), "java.lang.String");
	}

	@Test
	public void testFunctionConnector() {
		EPackage functionModelingPackage = Util.findEPackageByName("functionmodeling", generatedEcore);
		assertNotNull(functionModelingPackage);	
		assertTrue(functionModelingPackage.getESubpackages().size() == 1);
		EClass functionConnector = (EClass) functionModelingPackage.getEClassifier("FunctionConnector");
		assertNotNull(functionConnector);
		EReference port = (EReference) functionConnector.getEStructuralFeature("port");
		assertNotNull(port);
		assertEquals(port.getName(), "port");
		assertEquals(port.getUpperBound(), 2);
		assertEquals(port.getLowerBound(), 2);
	}
	
}
