/**
 * <copyright>
 * 
 * Copyright (c) Continental AG and others.
 * All rights reserved. This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 
 * which accompanies this distribution, and is
 * available at http://www.eclipse.org/org/documents/epl-v10.php
 * 
 * Contributors: 
 *     Continental AG - Initial API and implementation
 * 
 * </copyright>
 */
package eaadapter.xsd.output;

import java.util.List;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupDir;
import org.stringtemplate.v4.STGroupFile;

import eaadapter.xsd.model.EAXSDAbstract;
import eaadapter.xsd.model.EAXSDAdministrativeData;
import eaadapter.xsd.model.EAXSDAttribute;
import eaadapter.xsd.model.EAXSDCardinality;
import eaadapter.xsd.model.EAXSDClassType;
import eaadapter.xsd.model.EAXSDDataType;
import eaadapter.xsd.model.EAXSDDerived;
import eaadapter.xsd.model.EAXSDEnumeration;
import eaadapter.xsd.model.EAXSDGlobal;
import eaadapter.xsd.model.EAXSDModel;
import eaadapter.xsd.model.EAXSDReference;
import eaadapter.xsd.model.EAXSDSimpleAttribute;
import eaadapter.xsd.model.EAXSDSimpleType;
import eaadapter.xsd.model.EAXSDSubtypes;
import eaadapter.xsd.model.EAXSDType;

public class StringTemplateController {
    
    private STGroup group = null;
    private EAXSDModel model;
    private boolean minMax;

    public StringTemplateController(EAXSDModel root, String groupDir) {
        model = root;
        minMax = true;
        
        this.group = new STGroupDir(groupDir, '$', '$');
        this.group.registerRenderer(EAXSDType.class, new EAXSDTypeRenderer(
                root.getAdministrative().getNamespaceprefix() , true, true));
        this.group.registerRenderer(EAXSDCardinality.class, 
                new EAXSDCardinalityRenderer());
        
        STGroup utilities = new STGroupFile(groupDir + "/utilities.stg", '$', '$');
        utilities.registerRenderer(EAXSDCardinality.class, 
                new EAXSDCardinalityRenderer());
        this.group.importTemplates(utilities);
    }

    public void setMinMax(boolean minMax) {
        this.minMax = minMax;
    }

    protected ST getStringTemplate(String name) {
        
        ST st = this.group.getInstanceOf(name);
       
        st.add("nsuri", model.getAdministrative()
                .getNamespaceuri());
        st.add("nsprefix", model.getAdministrative()
                .getNamespaceprefix());
        st.add("minmax", Boolean.valueOf(minMax));
        
        return st;
    }
    
    public String getGlobalXSDString(EAXSDGlobal global) {
        
        ST st = getStringTemplate("global");
        st.add("global", global);
        
        return st.render();
    }

    public String getElementCaseXSDString(int caseno, EAXSDDataType xsdcase) {

        ST st = getStringTemplate(String.format("elementcase%02d",
                Integer.valueOf(caseno)));
        st.add("case", xsdcase);

        return st.render();
    }

    public String getDisclaimerXSDString(String disclaimer) {
        
        ST st = getStringTemplate("disclaimer");
        st.add("disclaimer", disclaimer);
        
        return st.render();
    }

    public String getSkeletonXSDString(EAXSDAdministrativeData admindata, String disclaimer,
            List<String> elements) {
        
        ST st = getStringTemplate("skeleton");
        st.add("elements", elements);
        st.add("disclaimer", disclaimer);
        st.add("admindata", admindata);
        
        return st.render();
    }
    
    public String getDerivedXSDString(EAXSDDerived derived) {

        ST st = getStringTemplate("derived");
        st.add("derived", derived);

        return st.render();
    }
    
    public String getAbstractTypeXSDString(EAXSDAbstract abstracttype) {
        
        ST st = getStringTemplate("abstracttype");
        st.add("abstracttype", abstracttype);
        
        return st.render();
    }

    public String getSimpleTypeXSDString(EAXSDSimpleType simpletype, String attributegroup) {
        
        ST st = getStringTemplate("simpletype");
        st.add("attributegroup", attributegroup);
        st.add("simpletype", simpletype);
        
        
        return st.render();
    }

    public String getAttributeGroupXSDString(EAXSDClassType group, List<String> simpleattributes) {
        
        ST st = getStringTemplate("attributegroup");
        st.add("simpleattributes", simpleattributes);
        st.add("attributegroup", group);
        
        return st.render();
    }

    public String getAttributeGroupXSDString(EAXSDSimpleType simpletype,
            List<String> simpleattributes) {
        
        ST st = getStringTemplate("attributegroupOfSimpleType");
        st.add("simpleattributes", simpleattributes);
        st.add("simpletype", simpletype);
        
        return st.render();
    }
    
    public String getAttributeXSDString(EAXSDAttribute attribute) {
        
        ST st = getStringTemplate("attribute");
        st.add("attribute", attribute);
        
        return st.render();
    }

    public String getSimpleAttributeXSDString(EAXSDSimpleAttribute simpleattribute) {
        
        ST st = getStringTemplate("simpleattribute");
        st.add("simpleattribute", simpleattribute);
        
        return st.render();
    }
    
    public String getEnumerationXSDString(EAXSDEnumeration enumeration) {
        
        ST st = getStringTemplate("enumeration");
        st.add("enum", enumeration);
        
        return st.render();
    }

    public String getReferenceXSDString(EAXSDReference reference) {
        
        ST st = getStringTemplate("reference");
        st.add("reference", reference);
        
        return st.render();
    }

    public String getSubTypesXSDString(EAXSDSubtypes subtypes) {
        
        ST st = getStringTemplate("subtypes");
        st.add("subtypes", subtypes);
        
        return st.render();
    }
    
    public String getClassTypeXSDString(EAXSDClassType classtype, String attributegroup,
            List<String> elementattributes, String abstracttype,
            String subtypes, String global) {
        
        ST st = getStringTemplate("classtype");
        st.add("elementattributes", elementattributes);
        st.add("global", global);
        st.add("classtype", classtype);
        st.add("subtypes", subtypes);
        st.add("attributegroup", attributegroup);
        st.add("abstracttype", abstracttype);
        
        return st.render();
    }
}
